package reksten.base;

import java.util.Arrays;

public class QFDistance implements Distance<FloatVector> {

	private int calculations;
	
	private float[][] A;
	private float EPSILON;
	
	public int maxiter = 0;
	
	public QFDistance(float[][] A, float EPSILON) {
		if(A.length==0 || A.length!=A[0].length)
			throw new IllegalArgumentException("QFDistance needs a aquare matrix!");
		this.A = A;
//		if(!isPositiveDefinite())
//			throw new IllegalArgumentException("Matrix must be positive definite!");
		this.EPSILON = EPSILON;
	}
	
	public long calculations() {
		return this.calculations;
	}

	public float distance(FloatVector a, FloatVector b) {
		calculations++;
		float dist = 0.0f;
		
		float[] diff = new float[A.length];
		for(int i =0;i<A.length;i++)
			diff[i] = (float)Math.abs(a.get(i)-b.get(i));
		
		for(int i = 0;i<A.length;i++) {
			for(int j = 0;j<A.length;j++) {
				dist += A[i][j] * diff[i] * diff[j];
			}
		}
		
		return (float)Math.sqrt(dist);
	}
	
	public float distance(float[] diff) {
		calculations++;
		float dist = 0.0f;
		
		for(int i = 0;i<A.length;i++)
			for(int j = 0;j<A.length;j++)
				dist += A[i][j] * diff[i] * diff[j];
		
		return (float)Math.sqrt(dist);
	}
	
	public float boxDistance(FloatVector query, float[] box_low, float[] box_upp, float bound) {
		int N = A.length;
		// 1. Move query object to origo (box follows)
		float[] point = new float[N], bl = new float[N], bu = new float[N];
		for(int i = 0;i<N;i++)point[i] = query.get(i);
		System.arraycopy(box_low,0,bl,0,N);
		System.arraycopy(box_upp,0,bu,0,N);
		for(int i = 0;i<N;i++) {
			bl[i] -= point[i];
			bu[i] -= point[i];
		}
		Arrays.fill(point, 0);
		
		float distance = Float.MAX_VALUE;
		// 2. Loop until no improvement on distance
		int iterations = 0;
		while(true) {
			iterations++;
			
			//		a. Set point to its closest point in the box.
			closest(point,bl,bu);
			
			//		b. Update distance.
			float ndist = distance(point);
			
			if(ndist<bound)
				return ndist;
			
//			if(iterations%10==0)
//				System.err.println(iterations+" "+ndist+" "+(distance-EPSILON));
			
			//		c. If improvement was zero, break.
			if(ndist>=distance-EPSILON)break;
			if(ndist<EPSILON)return 0.0f;
			distance = ndist;
			
			//		d. Find descending gradient of ellipse in point.
			float[] descendinggrad = descendingGradient(point);
			float[] gradient = new float[N];
			System.arraycopy(descendinggrad,0,gradient,0,N);
			
			//		e. Change gradient to not exit the box
			//		f. If gradient is zero, break.
			if(truncateGradient(gradient, point, bl, bu))
				break;
			
			//		g. Linear minimization from p in direction of g.
			float scale = linearminimization(descendinggrad, gradient);
			
			//		h. Set point to (point + linear minimization * gradient)
			for(int i = 0;i<N;i++)point[i] += scale * gradient[i];
			
		}
		maxiter = Math.max(maxiter, iterations);
		return distance;
		
	}
	
	private void closest(float[] p, float[] bl, float[] bu) {
		for(int i = 0;i<p.length;i++) {
			if(p[i]<bl[i])p[i] = bl[i];
			if(p[i]>bu[i])p[i] = bu[i];
		}
	}
	
	// Returns true if all dimensions are truncated!
	private boolean truncateGradient(float[] gradient, float[] p, float[] bl, float[] bu) {
		boolean trunc = true;
		for(int i = 0;gradient.length>i;i++) {
			if(gradient[i]<0.0f && p[i]<bl[i]+Helper.EPSILON)
				gradient[i] = 0.0f;
			if(gradient[i]>0.0f && p[i]>bu[i]-Helper.EPSILON)
				gradient[i] = 0.0f;
			if(!Helper.equals(0.0, gradient[i]))
				trunc = false;
		}
		return trunc;
	}
	
	private float linearminimization(float[] pgrad, float[] g) {
		float num = 0.0f;
		float den = 0.0f;
		float[] ggrad = descendingGradient(g);
		for(int i = 0;i<pgrad.length;i++) {
			num -= pgrad[i]*g[i];
			den += ggrad[i]*g[i];
		}
		return num/den;
	}
	
	private float[] descendingGradient(float[] p) {
		calculations++;
		float[] grad = new float[p.length];
		for(int i = 0;i<p.length;i++)
			for(int j = 0;j<p.length;j++)
				grad[i] -= A[i][j] * p[j];
		return grad;
	}
	
	public boolean isPositiveDefinite() {
//		System.err.println("GO");
		for(int size = 1;size<=A.length;size++) {
			
			float[][] mtrx = new float[size][size];
			for(int i = 0;i<size;i++) {
				for(int j = 0;j<size;j++) {
					mtrx[i][j] = A[i][j];
				}
			}
			double det = determinant(mtrx);
//			System.err.println(det);
			if(det<-EPSILON)
				return false;
		}
		return true;
	}
    
    public static float determinant(float[][] A) {
		
		if(A.length!=A[0].length)throw new IllegalArgumentException("Size mismatch!");
		
		// Lag kopi av matrisen:
		float[][] aug = new float[A.length][A.length];
		for(int i = 0;i<aug.length;i++)
			System.arraycopy(A,0,aug,0,A.length);
		
		// To echelon form:
		int var = 0;

		float determinant = 1.0f;
		for(int i = 0;i<A.length && var<A[0].length;var++) {
			// Find a row with nonzero var variable:
			int x = i;
			while(x<A.length && zero(aug[x][var]))x++;
			if(x==A.length)continue;
			swap(aug, x, i);
			if(x!=i)determinant = -determinant;
				
			// scale row:
//			for(int j = aug[i].length-1;j>=var;j--)
//				aug[i][j] /= aug[i][var];
			
			// gauss-eliminate downwards:
			for(int j = i+1;j<A.length;j++) {
				if(zero(aug[j][var]))continue;
				for(int k = aug[j].length-1;k>=var;k--)
					aug[j][k] -= aug[j][var]*aug[i][k];
			}
			i++;
		}

		for(int i = 0;i<A.length;i++) {
			determinant *= A[i][i];
		}
		
//		for(float[] arr : aug)System.err.println(Arrays.toString(arr));
		
		
		return determinant;
		
	}
    
    public static boolean zero(double val) {
		return Math.abs(val)<10e-10;
	}
	
	public static void swap(float[][] A, int ind1,int ind2) {
		float[] temp = A[ind1];
		A[ind1] = A[ind2];
		A[ind2] = temp;
	}

}
