package reksten.linearscan;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.PriorityQueue;

import reksten.base.DataObject;
import reksten.base.Distance;
import reksten.base.ResultEntry;
import reksten.base.SimSearchIndex;

import neustore.base.DBBuffer;
import neustore.base.DBBufferReturnElement;
import neustore.base.DBIndex;

/**
 * Linear Scan Index
 * Can be used for comparison when evaluating other indices. 
 * 
 * @author eireksten
 *
 * @param <D>
 */

public class LScan<D extends DataObject> extends DBIndex implements SimSearchIndex<D> {
	
//	private boolean DEBUG = false;
	
	private int lastpage;
	private Distance<D> distance;
	private int size = 0;
	
	public LScan(DBBuffer _buffer, String filename, boolean isCreate, Distance<D> dist) throws IOException {
        super(_buffer, filename, isCreate);
        this.distance = dist;
        this.size = 0;
    }
	
//	public void setDebug(boolean debug) {
//		this.DEBUG = debug;
//	}
	
	public int size() {
		return this.size;
	}
	
	/**
	 * Inserts a new object into the data base.
	 * 
	 * @param data The object to be inserted
	 * @throws IOException
	 */
	public void insert(D data) {
		try {
			size++;
			LScanPage<D> last = readPage(lastpage);
			if(!last.enoughSpace(data)) {
				// Create a new page and set to last!
				last = new LScanPage<D>(LScanPage.DATA_PAGE, pageSize, lastpage);
				lastpage = allocate();
			}
			last.insert(data);
			buffer.writePage(file, lastpage, last);
		} catch(IOException e) {
			e.printStackTrace();
			throw new IllegalStateException("Unable to write page!");
		}
	}
	
	/**
	 * Makes a linear scan for the query object
	 * 
	 * @param data Query Object
	 * @return The data base object most similar to the query object. null if the database is empty.
	 * @throws IOException
	 */
	public List<D> rangeSearch(D query, float range) {
		List<D> result = new ArrayList<D>();
		int pageid = lastpage;
		while(pageid>=0) {
			LScanPage<D> page;
			try {
				page = readPage(pageid);
			} catch (IOException e) {
				e.printStackTrace();
				throw new IllegalStateException("Unable to read page!");
			}
			Iterator<D> it = page.iterator();
			while(it.hasNext()) {
				D next = it.next();
				float dist = distance.distance(query,next);
				if(dist<=range)result.add(next);
			}
			pageid = page.previous();
		}
		return result;
	}
	
	/**
	 * Makes a linear scan for the query object
	 * 
	 * @param data Query Object
	 * @param dist Distance measurement to be used
	 * @return The data base object most similar to the query object. null if the database is empty.
	 * @throws IOException
	 */
	public List<D> kNearestNeighborSearch(D query, int K) {
		PriorityQueue<ResultEntry<D>> queue = new PriorityQueue<ResultEntry<D>>();
		int pageid = lastpage;
		while(pageid>=0) {
			LScanPage<D> page;
			try {
				page = readPage(pageid);
			} catch (IOException e) {
				e.printStackTrace();
				throw new IllegalStateException("Unable to read page!");
			}
			Iterator<D> it = page.iterator();
			while(it.hasNext()) {
				D next = it.next();
				ResultEntry<D> entry = new ResultEntry<D>(next, distance.distance(query,next));
				if(queue.size()<K || entry.compareTo(queue.peek())>0) {
					if(queue.size()==K)queue.poll();
					queue.add(entry);
				}
			}
			pageid = page.previous();
		}
		List<D> list = new ArrayList<D>();
		while(!queue.isEmpty())
			list.add(queue.poll().getObject());
		return list;
	}
	
	@SuppressWarnings("unchecked")
	private LScanPage<D> readPage(int pageid) throws IOException {
		
		DBBufferReturnElement ret = buffer.readPage(file, pageid);
		if(ret.parsed)return (LScanPage<D>)ret.object;
		
		byte[] page = (byte[])ret.object;
		LScanPage<D> parsedpage = new LScanPage(0,0,0);
        parsedpage.read(page);
		
        ret.object = parsedpage;
        ret.parsed = true;
        
		return parsedpage;
	}
	
	/**
	 * Initializes index head information.
	 * Called when creating a DBIndex.
	 */	
	@Override
	protected void initIndexHead() throws IOException {
		
		lastpage = allocate();
        LScanPage<D> page = new LScanPage<D>(LScanPage.DATA_PAGE, pageSize, LScanPage.NO_PAGE);
        buffer.writePage(file, lastpage, page);
        
	}

	/**
	 * Reads index head information from a byte array.
	 * Called when opening a DBIndex.
	 * @param indexHead   the byte array of length pageSize-OVERHEAD
	 */	
	@Override
	protected void readIndexHead(byte[] indexHead) throws IOException {
		
		ByteArrayInputStream byte_in = new ByteArrayInputStream(indexHead);
        DataInputStream in = new DataInputStream(byte_in);
        
        lastpage = in.readInt();
        
        in.close();
        byte_in.close();
	}

	/**
	 * Writes index head information to a byte array.
	 * Called before closing a DBIndex.
	 * @param indexHead   the byte array of length pageSize-OVERHEAD
	 */
	@Override
	protected void writeIndexHead(byte[] indexHead) throws IOException {
		
		ByteArrayOutputStream byte_out = new ByteArrayOutputStream(pageSize);
		DataOutputStream out = new DataOutputStream(byte_out);
		
		out.writeInt(this.lastpage);
		
		byte[] result = byte_out.toByteArray();
		System.arraycopy(result, 0, indexHead, 0, result.length );
		
		out.close();
		byte_out.close();
	}

}
