package reksten.pmtree;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.PriorityQueue;
import java.util.Stack;

import reksten.base.DataObject;
import reksten.base.Distance;
import reksten.base.ResultEntry;
import reksten.base.SimSearchIndex;

import neustore.base.ByteArray;
import neustore.base.DBBuffer;
import neustore.base.DBBufferReturnElement;
import neustore.base.DBIndex;

public class PMTree<Data extends DataObject> extends DBIndex implements SimSearchIndex<Data> {

	private boolean PTOLEMAIC = true;
	
	public int PAGE_SIZE = 8192;
	private int rootid;					// Page ID of root page
	private int size;
	private Distance<Data> distance;	// Distance calculation
	private List<Data> pivots;
	private float[][] interpivotdists;
	
	public PMTree(DBBuffer buffer, String filename, int PAGE_SIZE, List<Data> pivots, Distance<Data> measure, boolean ptolemaic) throws IOException {
		super(buffer, filename, true);
		this.PAGE_SIZE = PAGE_SIZE;
		this.pivots = pivots;
		this.distance = measure;
		int N = pivots.size();
		
		this.PTOLEMAIC = ptolemaic;
		
		if(PTOLEMAIC) {
			interpivotdists = new float[N][N]; // Needed for ptolemaic pruning!
			for(int i = 0;i<N;i++)
				for(int j = i+1;j<N;j++)
					interpivotdists[i][j] = interpivotdists[j][i] = distance.distance(pivots.get(i), pivots.get(j));
		}
		
		
	}
	
	public PMTree(DBBuffer buffer, String filename, int PAGE_SIZE, boolean ptolemaic) throws IOException {
		super(buffer, filename, false);
		this.PAGE_SIZE = PAGE_SIZE;
		this.PTOLEMAIC = ptolemaic;
	}
	
	public List<Data> kNearestNeighborSearch(Data query, int K) {
		
		// 1. Calculate the pivotdistances for the query object!
		float[] querypivotdistances = pivotdistances(query);
		
		PriorityQueue<ResultEntry<Data>> results = new PriorityQueue<ResultEntry<Data>>();
		
		Stack<Integer> idstack = new Stack<Integer>();
		Stack<Float> diststack = new Stack<Float>();
		idstack.push(rootid);
		diststack.push(-1.0f);
		
		while(!idstack.isEmpty()) {
			int nextpage = idstack.pop();
			float pardist = diststack.pop();
			PMTreePage<Data> page = getPage(nextpage);
			 
			if(page.isLeaf()) { // Leaf Node
				Iterator<PMTreeEntry<Data>> iter = page.iterator(); 
				while(iter.hasNext()) {
					PMTreeLeafEntry<Data> entry = (PMTreeLeafEntry<Data>)iter.next();
					
					double range = Double.MAX_VALUE/2;
					if(results.size()==K)range = results.peek().distance();
					double upper_bound = Math.abs(entry.parentdistance - pardist);
					if(upper_bound<=range) { // Need to verify closer!
						
						boolean prune = false;
						for(int i = 0;i<querypivotdistances.length && !prune;i++)
							if(Math.abs(querypivotdistances[i]-entry.getPivotDistance(i))>range)
								prune = true;

						if(PTOLEMAIC) {
							for(int i = 0;i<interpivotdists.length && !prune;i++) {
								for(int j = 0;j<interpivotdists.length && !prune;j++) {
									if(i==j)continue;
									// Use Ptolemaic pruning with the pivots i and j!
									float ptolLB = (entry.getPivotDistance(i) * querypivotdistances[j] - entry.getPivotDistance(j) * querypivotdistances[i] ) / interpivotdists[i][j];
									if(ptolLB>range) {
										prune = true;
//										System.err.println("Leaf Prune!");
									}
								}
							}
						}
						if(!prune) {
							float truedist = this.distance(entry.object,query);
							if(truedist<=range) {
								if(results.size()==K)results.poll();
								results.add(new ResultEntry<Data>(entry.object,truedist));
							}
						}
					}
				}
			} else { // Internal Node
				Iterator<PMTreeEntry<Data>> iter = page.iterator(); 
				while(iter.hasNext()) {
					PMTreeInternalEntry<Data> entry = (PMTreeInternalEntry<Data>)iter.next();
					double range = Double.MAX_VALUE/2;
					if(results.size()==K)range = results.peek().distance();
					double upper_bound = Math.abs(entry.parentdistance - pardist) - entry.getRadius();
					if(upper_bound<=range) {
						boolean prune = false;
						for(int i = 0;i<querypivotdistances.length && !prune;i++) {
							float[] ring = entry.getHyperring(i);
							if(querypivotdistances[i]-range>ring[1] || querypivotdistances[i]+range<ring[0])
								prune = true;
							
						}
						if(PTOLEMAIC) {
							for(int i = 0;i<interpivotdists.length && !prune;i++) {
								for(int j = 0;j<interpivotdists.length && !prune;j++) {
									if(i==j)continue;
									// Use Ptolemaic pruning with the pivots i and j!
									float ptolLB = (entry.hyperrings_inner[i] * querypivotdistances[j] - entry.hyperrings_outer[j] * querypivotdistances[i] ) / interpivotdists[i][j];
									if(ptolLB>range) {
										prune = true;
//										System.err.println("Internal Prune!!");
									}
								}
							}
						}
						if(!prune) {
							float truedist = this.distance(entry.object, query);
							if(truedist<=range+entry.getRadius()) {
								idstack.push(entry.getSubtreeID());
								diststack.push(truedist);
							}
						}
					}
				}
			}
		}
		List<Data> ret = new ArrayList<Data>();
		while(!results.isEmpty())ret.add(results.poll().getObject());
		return ret;
	}
	
	public List<Data> rangeSearch(Data query, float range) {
		
		float[] querypivotdistances = pivotdistances(query);
		
		int nodesvisited = 0;
		List<Data> result = new LinkedList<Data>();
		
		Stack<Integer> idstack = new Stack<Integer>();
		Stack<Float> diststack = new Stack<Float>();
		idstack.push(rootid);
		diststack.push(-1.0f);
		
		while(!idstack.isEmpty()) {
			nodesvisited++;
			int nextpage = idstack.pop();
			float pardist = diststack.pop();
			PMTreePage<Data> page = getPage(nextpage);
			 
			if(page.isLeaf()) { // Leaf Node
				Iterator<PMTreeEntry<Data>> iter = page.iterator(); 
				while(iter.hasNext()) {
					PMTreeLeafEntry<Data> entry = (PMTreeLeafEntry<Data>)iter.next();
					double lower_bound = Math.abs(entry.parentdistance - pardist);
					if(lower_bound<=range) { // Need to verify closer!
						boolean prune = false;
						for(int i = 0;i<querypivotdistances.length && !prune;i++)
							if(Math.abs(querypivotdistances[i]-entry.getPivotDistance(i))>range)
								prune = true;
						if(PTOLEMAIC) {
							for(int i = 0;i<interpivotdists.length && !prune;i++) {
								for(int j = 0;j<interpivotdists.length && !prune;j++) {
									if(i==j)continue;
									// Use Ptolemaic pruning with the pivots i and j!
									float ptolLB = (entry.getPivotDistance(i) * querypivotdistances[j] - entry.getPivotDistance(j) * querypivotdistances[i] ) / interpivotdists[i][j];
									if(ptolLB>range) {
										prune = true;
//										System.err.println("Leaf Prune");
									}
								}
							}
						}
						if(!prune) {
							float truedist = this.distance(entry.object,query);
							if(truedist<=range)result.add(entry.object);
						}
						
					}
				}
			} else { // Internal Node
				Iterator<PMTreeEntry<Data>> iter = page.iterator(); 
				while(iter.hasNext()) {
					PMTreeInternalEntry<Data> entry = (PMTreeInternalEntry<Data>)iter.next();
					double upper_bound = Math.abs(entry.parentdistance - pardist) - entry.getRadius();
					if(upper_bound<=range) {
						boolean prune = false;
						for(int i = 0;i<querypivotdistances.length && !prune;i++) {
							float[] ring = entry.getHyperring(i);
							if(querypivotdistances[i]-range>ring[1] || querypivotdistances[i]+range<ring[0])
								prune = true;
							
						}
						if(PTOLEMAIC) {
							for(int i = 0;i<interpivotdists.length && !prune;i++) {
								for(int j = 0;j<interpivotdists.length && !prune;j++) {
									if(i==j)continue;
									// Use Ptolemaic pruning with the pivots i and j!
									float ptolLB = (entry.hyperrings_inner[i] * querypivotdistances[j] - entry.hyperrings_outer[j] * querypivotdistances[i] ) / interpivotdists[i][j];
									if(ptolLB>range) {
										prune = true;
//										System.err.println("Internal Prune!!");
									}
								}
							}
						}
						if(!prune) {
							float truedist = this.distance(entry.object, query);
							if(truedist<=range+entry.getRadius()) {
								idstack.push(entry.getSubtreeID());
								diststack.push(truedist);
							}
						}
					}
				}
			}
		}
		return result;
	}
	
	public void insert(Data object) {
//		System.err.println("INSERTION PMTREE!");
		size++;
		
		if(rootid==-1) { // Create the tree!
			PMTreePage<Data> nroot = new PMTreePage<Data>(PMTreePage.LEAF_NODE, PAGE_SIZE);
			nroot.insert(this,object,null);
			rootid = addPage(nroot);
			return;
		}
		
		PMTreePage<Data> root = getPage(rootid);
		boolean split = root.insert(this,object,null);
		
		if(split) {
			PMTreeInternalEntry<Data>[] newentries = split(root);
			PMTreePage<Data> nroot = new PMTreePage<Data>(PMTreePage.INTERNAL_NODE, PAGE_SIZE);
			nroot.addEntry(newentries[0]);
			nroot.addEntry(newentries[1]);
			rootid = addPage(nroot);
		} else {
			updatePage(rootid, root);
		}
		
	}
	
	@SuppressWarnings("unchecked")
	public PMTreeInternalEntry<Data>[] split(PMTreePage<Data> page) {
//		System.err.println("SPLIT PMTREE!");
		// 1. Choose two pivots! Use routing choice policy from paper.
		// Policy M_LB_DIST (Maximum Lower Bound on Distance)
		// First pivot is parent pivot
		Data pivotA = page.getRandomObject();
		Data pivotB = null;
		float bestdistance = 0.0f;
		for (int i = 0; i < page.elements.size(); i++) {
//			System.err.println("Pivot Choice Distance");
			float dist = this.distance(pivotA, page.elements.get(i).getObject());
			if (pivotB==null || dist > bestdistance) {
				pivotB = page.elements.get(i).getObject();
				bestdistance = dist;
			}
		}
		// 2. Distribute entries between the two pages! Use distribution policy from paper
		// Generalized Hyperplane partitioning!
		PMTreePage<Data> pageA = new PMTreePage<Data>(page.nodeType, page.pageSize);
		PMTreePage<Data> pageB = new PMTreePage<Data>(page.nodeType, page.pageSize);
		
		for (PMTreeEntry<Data> entry : page.elements) { // Insert the entry into the page with the closest pivot
//			System.err.println("Distribution Distance");
//			System.err.println("Distribution Distance");
			float distA = distance(pivotA, entry.getObject());
			float distB = distance(pivotB, entry.getObject());
			if (distA < distB) {
				entry.parentdistance = distA;
				pageA.addEntry(entry);
			} else {
				entry.parentdistance = distB;
				pageB.addEntry(entry);
			}
		}
		
		updatePage(rootid, pageA);
		int idB = addPage(pageB);
		
		PMTreeInternalEntry<Data> A = new PMTreeInternalEntry<Data>(this,pageA,pivotA,rootid);
		PMTreeInternalEntry<Data> B = new PMTreeInternalEntry<Data>(this,pageB,pivotB,idB);
		
		return new PMTreeInternalEntry[] {A, B};

	}
	
	public int size() {
		return this.size;
	}
	
	public float distance(Data a, Data b) {
		if(a==null || b==null)return 0.0f;
		return distance.distance(a, b);
	}
	
	public long distanceCalculations() {
		return this.distance.calculations();
	}
	
	public float[] pivotdistances(Data object) {
		float[] list = new float[pivots.size()];
		for(int i = 0;i<pivots.size();i++)
			list[i] = distance(pivots.get(i),object);
		return list;
	}
	
	public void updatePage(int id, PMTreePage<Data> page) {
		try {
			buffer.writePage(file, id, page);
		} catch(IOException e) {
			throw new IllegalStateException("Unable to write to file!");
		}
	}
	
	public int addPage(PMTreePage<Data> page) {
		int id = allocate();
		try {
			buffer.writePage(file, id, page);
		} catch(IOException e) {
			throw new IllegalStateException("Unable to write to file!");
		}
		return id;
	}
	
	@SuppressWarnings("unchecked")
	public PMTreePage<Data> getPage(int pageid) {
		try {
			DBBufferReturnElement ret = buffer.readPage(file, pageid);
			PMTreePage<Data> thepage = null;
			if(ret.parsed) {
				thepage = (PMTreePage<Data>) ret.object;
			} else {
				byte[] page = (byte[]) ret.object;
	            thepage = new PMTreePage<Data>(0,0);
	            thepage.read(page);
			}
			return thepage;
		} catch(IOException e) {
        	throw new IllegalStateException("Unable to read page!");
		}
		
	}

	@Override
	protected void initIndexHead() throws IOException {
		this.rootid = -1;
		this.size = 0;
	}

	@Override
	protected void readIndexHead(byte[] indexHead) throws IOException {
		// TODO Add storing of pivotlist!
		ByteArray array = new ByteArray(indexHead, true);
		rootid = array.readInt();
		size = array.readInt();
	}

	@Override
	protected void writeIndexHead(byte[] indexHead) throws IOException {
		// TODO Add storing of pivotlist!
		ByteArray array = new ByteArray(indexHead, false);
		array.writeInt(rootid);
		array.writeInt(size);
	}

}
