package reksten.pmtree;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import reksten.base.DataObject;

import neustore.base.DBPage;

public class PMTreePage<Data extends DataObject> extends DBPage {

	static final int INTERNAL_NODE = 1;
	static final int LEAF_NODE = 2;

	static final int OVERHEAD = 16;

	protected List<PMTreeEntry<Data>> elements;
	private int usedspace = 0;

	public PMTreePage(int nodeType, int pageSize) {
		super(nodeType, pageSize);
		usedspace = OVERHEAD; // nodeType+pageSize+element.size
		elements = new ArrayList<PMTreeEntry<Data>>();
	}

	public int freeBytes() {
		return this.pageSize - usedspace;
	}

	public boolean isLeaf() {
		return this.nodeType == PMTreePage.LEAF_NODE;
	}
	
	public Iterator<PMTreeEntry<Data>> iterator() {
		return elements.iterator();
	}

	public Data getRandomObject() {
		return elements.get((int) (Math.random() * elements.size())).getObject();
	}

	public float calculateRadius(PMTree<Data> tree, Data pivot) {
		float radius = 0.0f;
		for (PMTreeEntry<Data> entry : elements)
			radius = Math.max(radius, tree.distance(pivot, entry.getObject()) + entry.getRadius());
		return radius;
	}

	public void updateHyperrings(PMTreeInternalEntry<Data> parententry) {
		for (PMTreeEntry<Data> ent : elements)
			ent.updateHyperRings(parententry);
	}

	public void addEntry(PMTreeEntry<Data> entry) {
		usedspace += entry.size();
		this.elements.add(entry);
	}
	
	private void removeEntry(int index) {
		PMTreeEntry<Data> entry = this.elements.get(index);
		usedspace -= entry.size();
		this.elements.remove(index);
	}

	// Inserts the object into the subtree, and returns false if a split of this page is necessary!
	public boolean insert(PMTree<Data> tree, Data object, Data parent) {

		if (this.nodeType == LEAF_NODE) {
			PMTreeLeafEntry<Data> groundEntry = new PMTreeLeafEntry<Data>(object, tree.distance(object, parent), tree.pivotdistances(object));
			this.addEntry(groundEntry);
		} else if (this.nodeType == INTERNAL_NODE) {
			
			// 1. Identify and insert into best subentry!
			float bestscore = Float.MAX_VALUE;
			int best = 0;
			for (int i = 0;i<elements.size();i++) {
				PMTreeInternalEntry<Data> internal = (PMTreeInternalEntry<Data>)elements.get(i);
				float score = tree.distance(object, internal.getObject()) - internal.getRadius();
				if (score < 0)
					score = -(score + internal.getRadius());
				if ((bestscore > 0.0f && score < bestscore) || (bestscore < 0 && score < 0 && score > bestscore)) {
					bestscore = score;
					best = i;
				}
			}
			
			PMTreeInternalEntry<Data> bestentry = (PMTreeInternalEntry<Data>)elements.get(best);
			PMTreePage<Data> bestpage = tree.getPage(bestentry.getSubtreeID());
			boolean tosplit = bestpage.insert(tree, object, ((PMTreeInternalEntry<Data>)elements.get(best)).getObject());
			
			// 2. Update page! Split if necessary!
			if(tosplit) {
				this.removeEntry(best);
				PMTreeInternalEntry<Data>[] newentries = split(tree,bestentry,bestpage);
				newentries[0].parentdistance = tree.distance(parent,newentries[0].getObject());
				newentries[1].parentdistance = tree.distance(parent,newentries[1].getObject());
				this.addEntry(newentries[0]);
				this.addEntry(newentries[1]);
			} else {
				bestentry.update(tree,bestpage);
				tree.updatePage(bestentry.getSubtreeID(), bestpage);
			}

		} else {
			throw new IllegalStateException("Insertion not supported for this node type (" + this.nodeType + ")");
		}
		
		if(this.freeBytes()<0)return true; // Split needed!
		return false;

	}

	@SuppressWarnings("unchecked")
	public PMTreeInternalEntry<Data>[] split(PMTree<Data> tree, PMTreeInternalEntry<Data> parententry, PMTreePage<Data> page) {

		// 1. Choose two pivots! Use routing choice policy from paper.
		// Policy M_LB_DIST (Maximum Lower Bound on Distance)
		// First pivot is parent pivot
		Data pivotA = parententry.getObject();
		Data pivotB = null;
		float bestdistance = 0.0f;
		
		for (int i = 0; i < page.elements.size(); i++) {
			if (pivotB==null || page.elements.get(i).getParentDistance() > bestdistance) {
				pivotB = page.elements.get(i).getObject();
				bestdistance = page.elements.get(i).getParentDistance();
			}
		}			
		
		// 2. Distribute entries between the two pages! Use distribution policy from paper
		// Generalized Hyperplane partitioning!
		PMTreePage<Data> pageA = new PMTreePage<Data>(page.nodeType, page.pageSize);
		PMTreePage<Data> pageB = new PMTreePage<Data>(page.nodeType, page.pageSize);
		
		for (PMTreeEntry<Data> entry : page.elements) { // Insert the entry into the page with the closest pivot
			float distA = entry.getParentDistance();
			float distB = tree.distance(pivotB, entry.getObject());
			if (distA < distB){pageA.addEntry(entry);}
			else {
				pageB.addEntry(entry);
				entry.parentdistance = distB;
			}
		}
		int idA = parententry.getSubtreeID();
		tree.updatePage(idA, pageA);
		int idB = tree.addPage(pageB);
		
		PMTreeInternalEntry<Data> A = new PMTreeInternalEntry<Data>(tree,pageA,pivotA,idA);
		PMTreeInternalEntry<Data> B = new PMTreeInternalEntry<Data>(tree,pageB,pivotB,idB);
		
		return new PMTreeInternalEntry[] {A, B};

	}

	@SuppressWarnings("unchecked")
	protected void read(byte[] page) throws IOException {

		DataInputStream in = new DataInputStream(new ByteArrayInputStream(page));

		this.pageSize = in.readInt();
		this.nodeType = in.readInt();
		this.usedspace = in.readInt();
		int len = in.readInt();
		for (int i = 0; i < len; i++)
			elements.add((PMTreeEntry<Data>) PMTreeEntry.parse(in));

	}

	protected void write(byte[] page) throws IOException {
		ByteArrayOutputStream byte_out = new ByteArrayOutputStream(pageSize);
		DataOutputStream out = new DataOutputStream(byte_out);

		out.writeInt(pageSize);
		out.writeInt(nodeType);
		out.writeInt(usedspace);
		out.writeInt(elements.size());
		for (PMTreeEntry<Data> el : elements)
			el.write(out);

		
		byte[] buffer = byte_out.toByteArray();
		if(this.usedspace!=buffer.length)System.err.println("Strange: "+usedspace+" "+buffer.length);
		System.arraycopy(buffer, 0, page, 0, buffer.length);
	}

}
