package reksten.xtree;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.PriorityQueue;
import java.util.Stack;

import reksten.base.Distance;
import reksten.base.FloatVector;
import reksten.base.QFDistance;
import reksten.base.ResultEntry;
import reksten.base.SimSearchIndex;

import neustore.base.ByteArray;
import neustore.base.DBBuffer;
import neustore.base.DBBufferReturnElement;
import neustore.base.DBIndex;

public class XTree<Data extends FloatVector> extends DBIndex implements SimSearchIndex<Data> {

	private int rootid = -1; // ID of root page!
	private int size;
	private Distance<Data> distance;	// Distance calculation
	private int pagesize;
	
	private double MAX_OVERLAP, MIN_FANOUT;
	
	public XTree(DBBuffer buffer, String filename, boolean isCreate, int pageSize, double MIN_FANOUT, double MAX_OVERLAP, Distance<Data> distance) throws IOException {
		super(buffer, filename, isCreate);
		this.pagesize = pageSize;
		this.distance = distance;
		this.MAX_OVERLAP = MAX_OVERLAP;
		this.MIN_FANOUT = MIN_FANOUT;
	}
	
	public List<Data> kNearestNeighborSearch(Data query, int K) {
		
		PriorityQueue<ResultEntry<Data>> results = new PriorityQueue<ResultEntry<Data>>();
		
		Stack<Integer> idstack = new Stack<Integer>();
		idstack.push(rootid);
		
		while(!idstack.isEmpty()) {
			int nextpage = idstack.pop();
			XTreePage<Data> page = getPage(nextpage);
			 
			if(page.isLeaf()) { // Leaf Node
				Iterator<XTreeEntry<Data>> iter = page.iterator(this); 
				while(iter.hasNext()) {
					XTreeLeafEntry<Data> entry = (XTreeLeafEntry<Data>)iter.next();
					
					float dist = distance.distance(query,entry.getObject());
					double range = Double.MAX_VALUE;
					if(results.size()==K)range = results.peek().distance();
					if(dist<range) {
						if(results.size()==K)results.poll();
						results.add(new ResultEntry<Data>(entry.getObject(),dist));
					}
				}
			} else { // Internal Node
				Iterator<XTreeEntry<Data>> iter = page.iterator(this); 
				while(iter.hasNext()) {
					XTreeDirectoryEntry<Data> entry = (XTreeDirectoryEntry<Data>)iter.next();
					
					float range = Float.MAX_VALUE;
					if(results.size()==K)range = results.peek().distance();
					
					double lower_bound = Double.MAX_VALUE;
					if(distance instanceof QFDistance) {
						lower_bound = ((QFDistance)distance).boxDistance(query, entry.mbr_low, entry.mbr_high, range);
					} else {
						lower_bound = entry.LB_distance(query);
					}
					
					if(lower_bound<range)idstack.push(entry.getSubtreeID());
					
				}
			}
		}
		List<Data> ret = new ArrayList<Data>();
		while(!results.isEmpty())ret.add(results.poll().getObject());
		return ret;
	}
	
	public List<Data> rangeSearch(Data query, float range) {
		
		List<Data> ret = new ArrayList<Data>();
		
		Stack<Integer> idstack = new Stack<Integer>();
		idstack.push(rootid);
		
		while(!idstack.isEmpty()) {
			int nextpage = idstack.pop();
			XTreePage<Data> page = getPage(nextpage);
			 
			if(page.isLeaf()) { // Leaf Node
				Iterator<XTreeEntry<Data>> iter = page.iterator(this); 
				while(iter.hasNext()) {
					XTreeLeafEntry<Data> entry = (XTreeLeafEntry<Data>)iter.next();
					
					float dist = distance.distance(query,entry.getObject());
					
					if(dist<=range)ret.add(entry.getObject());
				}
			} else { // Internal Node
				Iterator<XTreeEntry<Data>> iter = page.iterator(this); 
				while(iter.hasNext()) {
					XTreeDirectoryEntry<Data> entry = (XTreeDirectoryEntry<Data>)iter.next();
					
					double lower_bound = Double.MAX_VALUE;
					if(distance instanceof QFDistance) {
						lower_bound = ((QFDistance)distance).boxDistance(query, entry.mbr_low, entry.mbr_high, range);
					} else {
						lower_bound = entry.LB_distance(query);
					}
					
					if(lower_bound<=range)idstack.push(entry.getSubtreeID());
					
				}
			}
		}
		return ret;
	}
	
	public void insert(Data object) throws IOException {
		size++;
		
		if(rootid<0) { // Page does not exist!
			XTreePage<Data> nroot = new XTreePage<Data>(XTreePage.LEAF_NODE,this.pagesize);
			nroot.insert(this, object);
			this.rootid = this.addPage(nroot);
			return;
		}
		
		
		XTreePage<Data> rootpage = getPage(rootid);
		boolean overflowed = rootpage.insert(this,object);
		
		if(overflowed) {
			freePage(rootid);
			
			XTreeDirectoryEntry<Data>[] split = rootpage.split(this,rootpage,0);
			
			if(split.length>1) {
				split[0].setSplitHistory(1<<rootpage.LAST_SPLIT_AXIS);
				split[1].setSplitHistory(1<<rootpage.LAST_SPLIT_AXIS);
				XTreePage<Data> nroot = new XTreePage<Data>(XTreePage.DIRECTORY_NODE,this.pagesize);
				nroot.addEntry(this,split[0],false);
				nroot.addEntry(this,split[1],false);
				this.rootid = addPage(nroot);
			}
		} else {
			updatePage(rootid, rootpage);
		}
		
	}
	
	public double MAX_OVERLAP() {
		return this.MAX_OVERLAP;
	}
	
	public double MIN_FANOUT() {
		return this.MIN_FANOUT;
	}
	
	public int size() {
		return this.size;
	}
	
	public float distance(Data a, Data b) {
		return distance.distance(a, b);
	}
	
	public long distanceCalculations() {
		return this.distance.calculations();
	}

	public void updatePage(int id, XTreePage<Data> page) {
		try {
			buffer.writePage(file, id, page);
		} catch(IOException e) {
			throw new IllegalStateException("Unable to write to file!");
		}
	}
	
	public int addPage(XTreePage<Data> page) {
		int id = allocate();
		try {
			buffer.writePage(file, id, page);
		} catch(IOException e) {
			throw new IllegalStateException("Unable to write to file!");
		}
		return id;
	}
	
	@SuppressWarnings("unchecked")
	public XTreePage<Data> getPage(int pageid) {
		if(pageid==0)
			throw new IllegalStateException("Illegal page ID!");
		try {
			DBBufferReturnElement ret = buffer.readPage(file, pageid);
			XTreePage<Data> thepage = null;
			if(ret.parsed) {
				thepage = (XTreePage<Data>) ret.object;
			} else {
				byte[] page = (byte[]) ret.object;
	            thepage = new XTreePage<Data>(0,0);
	            thepage.read(page);
			}
			return thepage;
		} catch(IOException e) {
        	throw new IllegalStateException("Unable to read page!");
		}
		
	}
	
	@Override
	protected void initIndexHead() throws IOException {
		rootid = -1;
		size = 0;
	}

	@Override
	protected void readIndexHead(byte[] indexHead) throws IOException {
		ByteArray array = new ByteArray(indexHead, true);
		rootid = array.readInt();
		size = array.readInt();
	}

	@Override
	protected void writeIndexHead(byte[] indexHead) throws IOException {
		ByteArray array = new ByteArray(indexHead, false);
		array.writeInt(rootid);
		array.writeInt(size);
	}

}
