package reksten.xtree;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Arrays;
import java.util.List;

import reksten.base.FloatVector;
import reksten.base.Helper;


public class XTreeDirectoryEntry<Data extends FloatVector> extends XTreeEntry<Data> implements Comparable<XTreeDirectoryEntry<Data>> {

	protected float[] mbr_low, mbr_high;
	private long splithistory;
	private int subtreeid;
	
	XTreeDirectoryEntry() {
		this.nodeType = XTreeEntry.NODE_DIRECTORY;
	}
	
	XTreeDirectoryEntry(XTree<Data> tree, XTreePage<Data> page) {
		this.nodeType = XTreeEntry.NODE_DIRECTORY;
		
		List<XTreeEntry<Data>> entries = page.allEntries(tree);
		
		int dimensions = entries.get(0).dimensionality();
		
		mbr_low = new float[dimensions]; 
		mbr_high = new float[dimensions];
		Arrays.fill(mbr_low, Float.MAX_VALUE);
		Arrays.fill(mbr_high,-Float.MAX_VALUE);
		
		for(int i = 0;i<entries.size();i++) {
			XTreeEntry<Data> subentry = entries.get(i);
			subentry.updateParentMBR(this);
		}
//		System.err.println(Arrays.toString(mbr_low));
//		System.err.println(Arrays.toString(mbr_high));
	}
	
	XTreeDirectoryEntry(List<XTreeEntry<Data>> entries) {
		this.nodeType = XTreeEntry.NODE_DIRECTORY;
		
		int dimensions = entries.get(0).dimensionality();
		mbr_low = new float[dimensions]; 
		mbr_high = new float[dimensions];
		Arrays.fill(mbr_low, Float.MAX_VALUE);
		Arrays.fill(mbr_high,-Float.MAX_VALUE);
		
		for(int i = 0;i<entries.size();i++) {
			XTreeEntry<Data> subentry = entries.get(i);
			subentry.updateParentMBR(this);
		}
	}
	
	public void setSplitHistory(long splithistory) {
		this.splithistory = splithistory;
	}
	
	public void setSubtree(int ID) {
		this.subtreeid = ID;
	}

	public int getSubtreeID() {
		return this.subtreeid;
	}
	
	public long splitHistory() {
		return this.splithistory;
	}
	
	@Override
	public int size() {
		int size = 2; 				// nodetype
		size += 4; 					// dimension count
		size += 2*4*mbr_low.length; // MBR
		size += 4;					// subtreeid
		size += 8; 					// split history
		return size;
	}
	
	public void updateSingleInsertion(Data object) {
		for(int i = 0;i<mbr_low.length;i++) {
			mbr_high[i] = Math.max(object.get(i),mbr_high[i]);
			mbr_low[i] = Math.min(object.get(i),mbr_low[i]);
		}
	}
	
	public double separation(XTreeDirectoryEntry<Data> other, int dimension) {
		if(this.mbr_low[dimension]>other.mbr_low[dimension])
			return this.mbr_low[dimension]-other.mbr_high[dimension];
		return other.mbr_low[dimension]-this.mbr_high[dimension];
	}
	
	public double length(XTreeDirectoryEntry<Data> other, int dimension) {
		return Math.max(mbr_high[dimension],other.mbr_high[dimension])
					- Math.min(mbr_low[dimension],other.mbr_low[dimension]);
	}
	
	public double unionarea(XTreeDirectoryEntry<Data> other) {
		double area = Math.max(mbr_high[0],other.mbr_high[0])-Math.min(mbr_low[0],other.mbr_low[0]);
		for(int i = 1;i<mbr_low.length;i++)
			area *= Math.max(mbr_high[i],other.mbr_high[i])-Math.min(mbr_low[i],other.mbr_low[i]);
		return area;
	}
	
	public double area() {
		double area = mbr_high[0]-mbr_low[0];
		for(int i = 1;i<mbr_low.length;i++)
			area *= mbr_high[i]-mbr_low[i];
		return area;
	}
	
	public double enlargement(Data object) {
		double ar = area();
		
		double narea = Math.max(object.get(0),mbr_high[0])-Math.min(object.get(0),mbr_low[0]);
		
		for(int i = 1;i<mbr_low.length;i++)
			narea *= Math.max(object.get(i),mbr_high[i])-Math.min(object.get(i),mbr_low[i]);
		
		return narea - ar;
		
	}
	
	public double overlap(XTreeDirectoryEntry<Data> entry) {
		// Intersection div union
		double intersection = Math.min(this.mbr_high[0],entry.mbr_high[0])-Math.max(this.mbr_low[0],entry.mbr_low[0]);
		double union = Math.max(this.mbr_high[0],entry.mbr_high[0])-Math.min(this.mbr_low[0],entry.mbr_low[0]);
		
		if(intersection<=0.0)return 0.0;
		
		for(int i = 1;i<mbr_low.length;i++) {
			double in = Math.min(this.mbr_high[i],entry.mbr_high[i])-Math.max(this.mbr_low[i],entry.mbr_low[i]);
			if(in<=0.0)return 0.0;
			
			intersection *= in;
			union *= Math.max(this.mbr_high[i],entry.mbr_high[i])-Math.min(this.mbr_low[i],entry.mbr_low[i]);
		}
		
		return intersection / union;
	}
	
	public double LB_distance(Data object) {
		double lb = 0.0;
		for(int i = 0;i<mbr_low.length;i++) {
			if(object.get(i)<mbr_low[i])
				lb = Math.max(lb,mbr_low[i]-object.get(i));
			if(object.get(i)>mbr_high[i])
				lb = Math.max(lb,object.get(i)-mbr_high[i]);
		}
		return lb;
	}
	
	@Override
	protected void read(DataInputStream in) throws IOException {
		
		int len = in.readInt();
		mbr_low = new float[len];
		mbr_high = new float[len];
		for(int i = 0;i<len;i++) {
			mbr_low[i] = in.readFloat();
			mbr_high[i] = in.readFloat();
		}
		this.subtreeid = in.readInt();
		this.splithistory = in.readLong();
		
	}
	
	@Override
	protected void write(DataOutputStream out) throws IOException {
		super.write(out);
		out.writeInt(mbr_low.length);
		for(int i = 0;i<mbr_low.length;i++) {
			out.writeFloat(mbr_low[i]);
			out.writeFloat(mbr_high[i]);
		}
		out.writeInt(subtreeid);
		out.writeLong(splithistory);
	}

	@Override
	public int dimensionality() {
		return this.mbr_high.length;
	}
	
	public double overlaps(XTreeDirectoryEntry<Data> other, int dimension) {
		float min = Math.max(mbr_low[dimension],other.mbr_low[dimension]);
		float max = Math.min(mbr_high[dimension],other.mbr_high[dimension]);
		return min<max?max-min:0.0;
	}

	public static int SORTING_DIMENSION = 0;
	public int compareTo(XTreeDirectoryEntry<Data> o) {
		if(mbr_low[SORTING_DIMENSION]<o.mbr_low[SORTING_DIMENSION]-Helper.EPSILON)
			return -1;
		if(mbr_low[SORTING_DIMENSION]>o.mbr_low[SORTING_DIMENSION]+Helper.EPSILON)
			return 1;
		if(mbr_high[SORTING_DIMENSION]<o.mbr_high[SORTING_DIMENSION])
			return -1;
		if(mbr_high[SORTING_DIMENSION]>o.mbr_high[SORTING_DIMENSION])
			return 1;
		return 0;
	}

	@Override
	public void updateParentMBR(XTreeDirectoryEntry<Data> parent) {
		for(int i = 0;i<dimensionality();i++) {
			parent.mbr_low[i] = Math.min(parent.mbr_low[i],this.mbr_low[i]);
			parent.mbr_high[i] = Math.max(parent.mbr_high[i],this.mbr_high[i]);
		}
	}
	
//	@Override
//	public boolean equals(XTreeEntry<Data> o) {
//		if(!(o instanceof XTreeDirectoryEntry)) {
////			System.err.println("its a leaf :/");
//			return false;
//		}
//		XTreeDirectoryEntry<Data> other = (XTreeDirectoryEntry<Data>)o;
//		if(this.nodeType!=other.nodeType) {
////			System.err.println("node type");
//			return false;
//		}
//			
//		if(this.mbr_low.length!=other.mbr_low.length) {
////			System.err.println("length!!!???");
//			return false;
//		}
//		for(int i = 0;i<mbr_low.length;i++) {
//			if(!Helper.equals(mbr_low[i],other.mbr_low[i])) {
////				System.err.println(mbr_low[i]+" "+other.mbr_low[i]);
//				return false;
//			}
//			if(!Helper.equals(mbr_high[i],other.mbr_high[i])) {
////				System.err.println(mbr_high[i]+" "+other.mbr_high[i]);
//				return false;
//			}
//		}
//		if(this.subtreeid!=other.subtreeid) {
////			System.err.println("subtree id");
//			return false;
//		}
//		if(this.splithistory!=other.splithistory) {
////			System.err.println("split history");
//			return false;
//		}
//		
//		return true;
//	}
	
//	// TODO Remove!
//	public boolean contains(XTreeEntry<Data> subentry, int depth) {
//		if(subentry instanceof XTreeLeafEntry) {
//			XTreeLeafEntry<Data> sub = (XTreeLeafEntry<Data>)subentry;
//			Data object = sub.getObject();
//			for(int i = 0;i<mbr_low.length;i++) {
//				if(mbr_low[i]>object.get(i)+Helper.EPSILON || mbr_high[i]<object.get(i)-Helper.EPSILON) {
//					System.err.println(mbr_low[i]+" <= "+object.get(i)+" <= "+mbr_high[i]);
//					return false;
//				}
//					
//			}
//		} else {
//			XTreeDirectoryEntry<Data> sub = (XTreeDirectoryEntry<Data>)subentry;
//			for(int i = 0;i<mbr_low.length;i++) {
//				if(mbr_low[i]>sub.mbr_low[i]+Helper.EPSILON || mbr_high[i]<sub.mbr_high[i]-Helper.EPSILON) {
//					System.err.println(mbr_low[i]+" <= "+sub.mbr_low[i]+" ... "+sub.mbr_high[i]+" <= "+mbr_high[i]);
//					return false;
//				}
//			}
//		}
//		return true;
//	}

//	@Override
//	public XTreeDirectoryEntry<Data> copy() {
//		
//		XTreeDirectoryEntry entry = new XTreeDirectoryEntry<FloatVector>();
//		entry.mbr_high = new float[this.mbr_high.length];
//		entry.mbr_low = new float[this.mbr_low.length];
//		System.arraycopy(this.mbr_low,0,entry.mbr_low,0,mbr_low.length);
//		System.arraycopy(this.mbr_high,0,entry.mbr_high,0,mbr_high.length);
//		
//		entry.nodeType = this.nodeType;
//		entry.splithistory = this.splithistory;
//		entry.subtreeid = this.subtreeid;
//		
//		return entry;
//	}
	
}
