package reksten.xtree;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import reksten.base.FloatVector;
import reksten.base.Helper;

import neustore.base.DBPage;

public class XTreePage<Data extends FloatVector> extends DBPage {
	
	static final int LEAF_NODE = 1;
	static final int DIRECTORY_NODE = 2; // For now, also representing supernodes.
	
	static final int OVERHEAD = 24;
	
	private List<XTreeEntry<Data>> entries;
	private int nextpage = -1; // If this is a supernode, points to the next page of the node.
	private int usedspace = 0;
	
	public XTreePage(int nodeType, int pageSize) {
		super(nodeType, pageSize);
		this.entries = new ArrayList<XTreeEntry<Data>>();
		this.usedspace = OVERHEAD;
	}
	
	public int fanout() {
		if(isSupernode())throw new IllegalStateException("Unable to retrieve fanout for supernodes!");
		return this.entries.size();
	}
	
	public Iterator<XTreeEntry<Data>> iterator(XTree<Data> tree) {
		return allEntries(tree).iterator();
	}
	
	public int freeBytes() {
		return this.pageSize-usedspace;
	}
	
	public boolean isSupernode() {
		return this.nextpage>=0;
	}
	
	public boolean isLeaf() {
		return this.nodeType==XTreePage.LEAF_NODE;
	}
	
	public List<XTreeEntry<Data>> allEntries(XTree<Data> tree) {
		List<XTreeEntry<Data>> ret = new ArrayList<XTreeEntry<Data>>();
		XTreePage<Data> atpage = this;
		while(true) {
			for(XTreeEntry<Data> entry : atpage.entries)ret.add(entry);
			if(atpage.nextpage==-1)break;
			atpage = tree.getPage(atpage.nextpage);
		}
		return ret;
	}
	
	// Adds the entry into the node. Can overflow unless supernode
	public void addEntry(XTree<Data> tree, XTreeEntry<Data> entry, boolean supernode) {
		
		if(this.freeBytes()>=entry.size()) {
			this.usedspace += entry.size();
			this.entries.add(entry);
			return;
		}
		
		if(isSupernode()) {
			// Send to next page!
			tree.getPage(this.nextpage).addEntry(tree,entry,true);
			tree.updatePage(this.nextpage,tree.getPage(this.nextpage));
			
		} else if(supernode) {
			// Create new page!
			XTreePage<Data> npage = new XTreePage<Data>(this.nodeType, this.pageSize);	
			npage.addEntry(tree,entry,true);
			this.nextpage = tree.addPage(npage);
		} else {
			// Overflow the node!
			this.usedspace += entry.size();
			this.entries.add(entry);
		}
		
	}
	
	private void removeEntry(XTree<Data> tree, int index) {
		
		if(index<this.entries.size()) {
			this.usedspace -= this.entries.get(index).size();
			this.entries.remove(index);
		} else if(isSupernode()) {
			tree.getPage(nextpage).removeEntry(tree,index-entries.size());
			tree.updatePage(this.nextpage,tree.getPage(this.nextpage));
		} else {
			throw new IllegalArgumentException("Unable to remove entry!");
		}
		
	}
	
	public boolean insert(XTree<Data> tree, Data object) throws IOException {
		
		if(this.nodeType==XTreePage.LEAF_NODE) {
			
			XTreeLeafEntry<Data> leafentry = new XTreeLeafEntry<Data>(object);
			this.addEntry(tree,leafentry,false);

		} else if(this.nodeType==XTreePage.DIRECTORY_NODE) {
			
			List<XTreeEntry<Data>> entrylist = allEntries(tree);
			
			// 1. Choose subtree to insert into [R-tree]. Find subentry with the least area enlargement 
			
			int best = 0;
			double bestenlargement = ((XTreeDirectoryEntry<Data>)entrylist.get(0)).enlargement(object);
			double bestarea = ((XTreeDirectoryEntry<Data>)entrylist.get(0)).area();
			
			for(int i = 1;i<entrylist.size();i++) {
				double enl = ((XTreeDirectoryEntry<Data>)entrylist.get(i)).enlargement(object);
				if(enl<bestenlargement) {
					best = i; bestarea = ((XTreeDirectoryEntry<Data>)entrylist.get(0)).area(); bestenlargement = enl;
				} else if(Helper.equals(enl,bestenlargement)) {
					double ar = ((XTreeDirectoryEntry<Data>)entrylist.get(0)).area();
					if(ar<bestarea) { best = i; bestarea = ar; bestenlargement = enl; }
				}
			}
			// 2. Insert into subtree.
			XTreeDirectoryEntry<Data> bestentry = ((XTreeDirectoryEntry<Data>)entrylist.get(best));
			XTreePage<Data> bestpage = tree.getPage(bestentry.getSubtreeID());
			boolean tosplit = bestpage.insert(tree, object);
			
			// Update page! Split if necessary
			if(tosplit) {
				removeEntry(tree,best);
				tree.freePage(bestentry.getSubtreeID());
				XTreeDirectoryEntry<Data>[] newentries = split(tree, bestpage, bestentry.splitHistory());
				for(XTreeDirectoryEntry<Data> entry : newentries)
					this.addEntry(tree, entry, false);
				
			} else {
				bestentry.updateSingleInsertion(object);
				tree.updatePage(bestentry.getSubtreeID(), bestpage);
			} 
			
		}
		
		if(this.freeBytes()<0)return true; // Overflow of the disk page!
		return false;
	}
	
	// The return list will contain two elements if the page was split, and one if the page became a supernode
	@SuppressWarnings("unchecked")
	protected XTreeDirectoryEntry<Data>[] split(XTree<Data> tree, XTreePage<Data> prevpage, long parentsplithistory) {
		if(prevpage.isSupernode())
			throw new IllegalArgumentException("Supernode sent to splitting!");
		
		List<XTreeEntry<Data>> entries = prevpage.allEntries(tree);
		XTreePage<Data>[] split = null;
		if(prevpage.isLeaf()) {
			// We are splitting a leaf node!
			List<XTreeLeafEntry<Data>> leafentries = new ArrayList<XTreeLeafEntry<Data>>();
			for(XTreeEntry<Data> entry : entries)leafentries.add((XTreeLeafEntry<Data>)entry);
			
			// Choose arbitrary split dimension and split there!
			int dimensions = leafentries.get(0).dimensionality();
			int splitdim = (int)(Math.random()*dimensions);

			this.LAST_SPLIT_AXIS = splitdim;
			
			// Sort list and divide into two balanced nodes!
			XTreeLeafEntry.SORTING_DIMENSION = splitdim;
			Collections.sort(leafentries);
			
			split = new XTreePage[2];
			split[0] = new XTreePage<Data>(XTreePage.LEAF_NODE, prevpage.pageSize);
			split[1] = new XTreePage<Data>(XTreePage.LEAF_NODE, prevpage.pageSize);
			for(int i = 0;i<leafentries.size();i++) {
				if(i<leafentries.size()/2)split[0].addEntry(tree, leafentries.get(i), false);
				else split[1].addEntry(tree, leafentries.get(i), false);
			}
			
		} else {
			// We are splitting a directory node!
			
			List<XTreeDirectoryEntry<Data>> directoryentries = new ArrayList<XTreeDirectoryEntry<Data>>();
			for(XTreeEntry<Data> entry : entries)directoryentries.add((XTreeDirectoryEntry<Data>)entry);
			
			int minfanout = (int)(directoryentries.size() * tree.MIN_FANOUT());
			
			// Try a topological split! [R-tree]
			split = topologicalSplit(tree, directoryentries, minfanout);
			
			// TODO Re-add! What does this mean?
			double overlap = new XTreeDirectoryEntry<Data>(tree,split[0]).overlap(new XTreeDirectoryEntry<Data>(tree,split[1]));
			if(overlap>tree.MAX_OVERLAP()) {
				split = overlapFreeSplit(tree, directoryentries);
//				if(split[0].fanout()<minfanout || split[1].fanout()<minfanout) {
//					System.err.println("CREATING SUPERNODE");
//					split = createSupernode(tree, directoryentries);
//				}
			}
			
		}
		
		if(split==null)
			throw new IllegalStateException("Unable to handle node splitting!");
		
		if(split.length==1) { // Supernode
			
			XTreeDirectoryEntry<Data> entry = new XTreeDirectoryEntry<Data>(entries);
			entry.setSplitHistory(parentsplithistory);
			entry.setSubtree(tree.addPage(split[0]));
			
			return new XTreeDirectoryEntry[] {entry};
			
		} else if(split.length==2) {
			
			XTreeDirectoryEntry<Data> a = new XTreeDirectoryEntry<Data>(tree,split[0]);
			XTreeDirectoryEntry<Data> b = new XTreeDirectoryEntry<Data>(tree,split[1]);
			a.setSplitHistory(parentsplithistory|(1<<LAST_SPLIT_AXIS));
			b.setSplitHistory(parentsplithistory|(1<<LAST_SPLIT_AXIS));
			
			a.setSubtree(tree.addPage(split[0]));
			b.setSubtree(tree.addPage(split[1]));
			return new XTreeDirectoryEntry[] {a,b};
		}
		
		throw new IllegalStateException("Split length mismatch!");
	}
	
	protected int LAST_SPLIT_AXIS = 0;
	
	@SuppressWarnings("unchecked")
	private XTreePage<Data>[] topologicalSplit(XTree<Data> tree, List<XTreeDirectoryEntry<Data>> entries, int minfanout) {
		
		if(minfanout==0)minfanout++;
		
		// 1. Pick two seeds
		// For each pair of entries:
		//   Calculate unionarea - areaA - areaB
		// Choose pair with largest value
		int len = entries.size();
		int a = 0, b = 1;
		double bestvalue = 0.0;
		for(int i = 0;i<len;i++) {
			for(int j = i+1;j<len;j++) {
				double val = entries.get(i).unionarea(entries.get(j)) - entries.get(i).area() - entries.get(j).area();
				if(val>bestvalue) {
					bestvalue = val;
					a = i;
					b = j;
				}
				
			}
		}
		
		// 2. Choose split axis
		// For each axis
		//   Calculate separation div length of enclosing rectangle
		// Choose axis with largest value.
		int dimensions = entries.get(0).dimensionality();
		int splitdimension = 0;
		bestvalue = -2.0;
		for(int dim = 0;dim<dimensions;dim++) {
			double sep = entries.get(a).separation(entries.get(b),dim);
			double length = entries.get(a).length(entries.get(b),dim);
			double val = sep / length;
			if(val>bestvalue) {
				bestvalue = val;
				splitdimension = dim;
			}
		}
		
		// 3. Choose split index
		// Sort the entries on the given axis
		// For each split (within fanout boundaries
		//   Find MBR of the two sides, and calculate the overlap value
		// Choose the index with the lowest overlap
		XTreeDirectoryEntry.SORTING_DIMENSION = splitdimension;
		Collections.sort(entries);
		bestvalue = Double.MAX_VALUE;
		int splitindex = 0;
		for(int index = minfanout;index<=len-minfanout;index++) {
			List<XTreeEntry<Data>> aside = new ArrayList<XTreeEntry<Data>>();
			List<XTreeEntry<Data>> bside = new ArrayList<XTreeEntry<Data>>();
			for(int i = 0;i<index;i++)aside.add(entries.get(i));
			for(int i = index;i<len;i++)bside.add(entries.get(i));
			
			XTreeDirectoryEntry<Data> aa = new XTreeDirectoryEntry<Data>(aside);
			XTreeDirectoryEntry<Data> bb = new XTreeDirectoryEntry<Data>(bside);
			
			double val = aa.overlap(bb);
			if(val<bestvalue) {
				splitindex = index;
				bestvalue = val;
			}
		}
		
		LAST_SPLIT_AXIS = splitdimension;
		
		XTreePage<Data> pageA = new XTreePage<Data>(DIRECTORY_NODE, this.pageSize);
		XTreePage<Data> pageB = new XTreePage<Data>(DIRECTORY_NODE, this.pageSize);
		for(int i = 0;i<splitindex;i++)pageA.addEntry(tree,entries.get(i),false);
		for(int i = splitindex;i<len;i++)pageB.addEntry(tree,entries.get(i),false);
		
		return new XTreePage[] {pageA, pageB};
	}
	
	@SuppressWarnings("unchecked")
	private XTreePage<Data>[] overlapFreeSplit(XTree<Data> tree, List<XTreeDirectoryEntry<Data>> entries) {

		long splitdimensions = -1;
		for(XTreeDirectoryEntry<Data> entry : entries)
			splitdimensions &= entry.splitHistory();
		
		int dimensions = entries.get(0).dimensionality();
		for(int dim = 0;dim<dimensions;dim++) {
			if((splitdimensions&(1<<dim))!=0) {
				// dim is a potential split dimension!

				XTreeDirectoryEntry.SORTING_DIMENSION = dim;
				Collections.sort(entries);
				
				int splitindex = -1;
				double overlap = Double.MAX_VALUE;
				for(int second = 1;second<entries.size();second++) {
					double ov = entries.get(second-1).overlaps(entries.get(second),dim);
					if(ov<overlap) {
						splitindex = second;
						overlap = ov;
					}
				}
				
				if(splitindex>=0) {
					// Perform the split and return it!
					XTreePage<Data>[] split = new XTreePage[2];
					split[0] = new XTreePage(this.nodeType,this.pageSize);
					split[1] = new XTreePage(this.nodeType,this.pageSize);
					for(int i = 0;i<entries.size();i++) {
						if(i<splitindex)split[0].addEntry(tree, entries.get(i), false);
						else split[1].addEntry(tree, entries.get(i), false);
					}
					LAST_SPLIT_AXIS = dim;
					return split;
				}
			}
		} 
		throw new IllegalStateException("No overlap-minimal split found!");
	}
	
	@SuppressWarnings("unchecked")
	private XTreePage<Data>[] createSupernode(XTree<Data> tree, List<XTreeDirectoryEntry<Data>> entries) {
		XTreePage<Data> supernode = new XTreePage<Data>(XTreePage.DIRECTORY_NODE, this.pageSize);
		
		for(XTreeDirectoryEntry<Data> entry : entries)
			supernode.addEntry(tree, entry, true);
		
		return new XTreePage[] {supernode};
	}

	@Override
	@SuppressWarnings("unchecked")
	protected void read(byte[] page) throws IOException {
		DataInputStream in = new DataInputStream(new ByteArrayInputStream(page));

		this.pageSize = in.readInt();
		this.nodeType = in.readInt();
		this.usedspace = in.readInt();
		this.nextpage = in.readInt();
		
		int len = in.readInt();
		for (int i = 0; i < len; i++)
			entries.add((XTreeEntry<Data>) XTreeEntry.parse(in));
		
	}

	@Override
	protected void write(byte[] page) throws IOException {
		
		ByteArrayOutputStream byte_out = new ByteArrayOutputStream(pageSize);
		DataOutputStream out = new DataOutputStream(byte_out);
		
		out.writeInt(pageSize);
		out.writeInt(nodeType);
		out.writeInt(usedspace);
		out.writeInt(nextpage);
		
		out.writeInt(entries.size());
		for (XTreeEntry<Data> el : entries)
			el.write(out);
		
		byte[] buffer = byte_out.toByteArray();
		System.arraycopy(buffer, 0, page, 0, buffer.length);
	}
	

}
